from gherkin.parser import Parser as GherkinParser
import six

from .exceptions import StepException, UnknownStepException
from . import utils

if six.PY3:
    from importlib import reload
    from importlib.machinery import SourceFileLoader
else:
    from imp import load_source, reload


class Feature:
    def __init__(self, feature_dict, tester):
        self.tester = tester

        self.name = feature_dict['name']
        self.tags = [tag['name'][1:] for tag in feature_dict['tags']]

        self.tests = [Test(test_dict, self)
                      for test_dict in feature_dict['children']]

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, str(self))


class Test:
    def __init__(self, test_dict, feature):
        self.feature = feature

        self.name = test_dict['name']
        self.tags = [tag['name'][1:] for tag in test_dict['tags']]

        self.steps = []
        step_type = 'given'
        for step_dict in test_dict['steps']:
            if step_dict['keyword'].lower().strip() == 'then':
                step_type = 'then'
            self.steps.append(Step(step_dict, step_type, self))

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, str(self))

    def loop(self, obj, steps, *args, **kwargs):
        for idx, step in enumerate(steps):
            try:
                obj = step(obj, *args, **kwargs)
            except StepException as explain:
                if step.step_type == 'then':
                    return False, explain
                else:
                    # failed conditional i.e. not relevant
                    return None, explain
            if step.loop:
                for o in obj:
                    o_result, o_explain = self.loop(
                        o, steps[idx + 1:], *args, **kwargs)
                    if not o_result:
                        return o_result, o_explain
                return True, None
        return True, None

    def __call__(self, obj, *args, **kwargs):
        def output(res, msg):
            if kwargs.get('bdd_verbose'):
                return res, msg
            return res

        context = dict(kwargs)
        if 'bdd_verbose' in context:
            del context['bdd_verbose']

        res, msg = self.loop(obj, self.steps, *args, **context)
        return output(res, msg)


class Step:
    def __init__(self, step_dict, step_type, test):
        def _find_matching_expr(line):
            for regex, fn, loop in test.feature.tester.store.values():
                r = regex.match(line)
                if r:
                    return fn, loop, r.groups()
            msg = 'I didn\'t understand "{}"'.format(line)
            raise UnknownStepException(msg)

        self.test = test

        self.text = step_dict['text']
        self.step_type = step_type
        match = _find_matching_expr(self.text)
        self.expr_fn, self.loop, self.expr_groups = match

    def __str__(self):
        return '{} {}'.format(self.step_type.title(), self.text)

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, str(self))

    def __call__(self, *args, **kwargs):
        if self.expr_groups:
            args = args + self.expr_groups
        return self.expr_fn(*args, **kwargs)


class BDDTester:
    def __init__(self, step_path):
        self._load_step_definitions(step_path)
        self.gherkinparser = GherkinParser()

    def _load_step_definitions(self, filepath):
        reload(utils)
        if six.PY3:
            SourceFileLoader('', filepath).load_module()
        else:
            load_source('', filepath)
        self.store = dict(utils.store)

    def load_feature(self, feature_filepath):
        with open(feature_filepath) as f:
            feature_txt = f.read()
        return self._gherkinify_feature(feature_txt)

    def _gherkinify_feature(self, feature_txt):
        feature_dict = self.gherkinparser.parse(feature_txt)['feature']
        return Feature(feature_dict, self)



import re


store = {}


def given(pattern, loop=False):
    return call(pattern, loop)


def then(pattern, loop=False):
    return call(pattern, loop)


def call(pattern, loop=False):
    def decorated(fn):
        store[pattern] = (re.compile(r'^{}$'.format(pattern)), fn, loop)
        return fn  # this isn't needed, but useful for testing
    return decorated



from os.path import abspath, dirname, join

import pytest

from bdd_tester import BDDTester
from bdd_tester.exceptions import UnknownStepException


EGGFUL_LIST = '''
SHOPPING LIST
=============

 * Bread
 * Butter
 * Milk
 * Eggs
'''


EGGLESS_LIST = '''
SHOPPING LIST
=============

 * Carrots
 * Tomatoes
 * Toothpaste
'''


EGG_LETTER = '''
Dear sir or madam,

I am writing to you about eggs.

Yours, Andy
'''


def test_succeeds():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    feature = tester.load_feature(join(fixture_path, 'feature.feature'))
    success, msg = feature.tests[0](EGGFUL_LIST, bdd_verbose=True)
    assert success is True
    assert msg is None

    success = feature.tests[0](EGGFUL_LIST)
    assert success is True


def test_not_relevant():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    feature = tester.load_feature(join(fixture_path, 'feature.feature'))
    success, msg = feature.tests[0](EGG_LETTER, bdd_verbose=True)
    assert success is None
    assert str(msg) == 'Not a shopping list'


def test_fails():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    feature = tester.load_feature(join(fixture_path, 'feature.feature'))
    success, msg = feature.tests[0](EGGLESS_LIST, bdd_verbose=True)
    assert success is False
    assert str(msg) == '"Eggs" not found'


def test_missing_step_definition():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    with pytest.raises(UnknownStepException):
        tester.load_feature(join(fixture_path, 'bad_feature.feature'))


def test_feature_repr():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    feature = tester.load_feature(join(fixture_path, 'feature.feature'))
    assert repr(feature) == '<Feature (Shopping list tests)>'


def test_tags_present():
    fixture_path = join(dirname(abspath(__file__)), 'fixtures')

    tester = BDDTester(join(fixture_path, 'steps.py'))
    feature = tester.load_feature(join(fixture_path, 'feature.feature'))
    assert feature.tags[0] == 'feature-tag'
    test = feature.tests[0]
    assert test.tags[0] == 'test-tag'
    assert test.feature.tags[0] == 'feature-tag'



from setuptools import setup, find_packages
from os.path import abspath, dirname, join


path = abspath(dirname(__file__))
with open(join(path, 'README.rst')) as f:
    readme = f.read()

test_deps = [
    'coveralls',
    'pytest',
    'pytest-cov',
]
extras = {'test': test_deps}

setup(
    name='bdd-tester',
    description='A very very basic BDD test runner',
    author='Andy Lulham',
    author_email='a.lulham@gmail.com',
    version='0.0.8',
    packages=find_packages(),
    scripts=[join('bin', 'bdd_tester')],
    license='MIT',
    long_description=readme,
    install_requires=[
        'gherkin-official',
        'six',
    ],
    setup_requires=['pytest-runner'],
    tests_require=test_deps,
    extras_require=extras,
)
